/*============================================================================
/ psionfs_file.c
/ © Michael Pieper 15. Oct. 2000
/     Last Change: 11. Nov. 2000
/============================================================================
/
/	This c-File includes all "File"-Functions for the PSION-FileSystem
/
/============================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*===========================================================================
/	We also need our own Headerfiles
/============================================================================*/

#include "psionfs.h"
#include "debug.h"
#include "psionfs_file.h"	/* Including all the File-Filesystem-Functions	*/
#include "psionfs_rfsv.h"	/* Including all PSION-Commands					*/

#include "p_file.h"
#include "mc_rfile.h"

/*===========================================================================
/	Create is the function to create a new file
/============================================================================*/

int	psionfs_create	(void *_ns, void *_node, const char *name,
					int omode, int perms, vnode_id *vnid, void **cookie) {

	dircookie		*dirCookie;
	char 			text[128];
	nspace 			*ns = (nspace *)_ns;
	vnode 			*node = (vnode *)_node;
	vnode 			*newNode;
	int				result = EINVAL;
	
	sprintf(text, "## fs_create vnod_id:%Ld, name:%s, omode:%x, perms:%d\n", node->id, name, omode, perms);
	debug_print(text);

	dirCookie = (dircookie *)malloc(sizeof(dircookie));
	if (dirCookie != NULL) {
		result = dirCookie->fd = open_node(ns, node, omode, name);
		dirCookie->buff = NULL;
		dirCookie->pos = dirCookie->read = 0;
			
		*cookie = (void*)dirCookie;
		if (result > 0) {
			newNode = search_vnode(ns, vnid, node->id, name, NULL);
			if (new_vnode(ns->id, *vnid, (void *)newNode) != B_NO_ERROR) {
				if (newNode != NULL) free(newNode);
				result = EINVAL;
			} else {
				notify_listener(B_ENTRY_CREATED, ns->id, node->id, 0, 0, name);	/* Lets tell the system, that there is a new folder!	*/
				result = B_NO_ERROR;
			}
		}
	}

	sprintf(text, "-- fs_create: filehandler %04x accepted result:%d\n", dirCookie->fd, result);
	debug_print(text);
	return result;
}

/*===========================================================================
/	Read the data to the file which is specified through the cookie
/============================================================================*/

int psionfs_read(void *_ns, void *_node, void *_cookie, off_t pos, void *buf, size_t *len)
{
	dircookie 		*dirc 	= (dircookie*)_cookie;
	nspace* 		ns 		= (nspace*)_ns;	
	vnode* 			node 	= (vnode *)_node;
	ssize_t			ret;

	char text[128];
	int result = B_NO_ERROR;

	sprintf(text, "## fs_read: vnod_id:%Ld, pos:%Ld, len:%lu %s\n", node->id, pos, *len, node->filename);
	debug_print(text);

	psion_fseek (ns, dirc, SEEK_SET, pos); 
	ret = psion_read (ns, dirc, (char *)buf, *len);

	if (ret > 0) *len = ret;									/* some Bytes read. everythings ok */
	else {
		*len = 0;												/* otherwise, set len to 0-Bytes read	*/
		switch(ret) {
			case B_NO_ERROR:	result = B_NO_ERROR;	break;	/* Nothing there to read, but no error (EOF)	*/
			default:			result = B_FILE_ERROR;	break;	/* otherwise general filerror					*/
		}
	}
	
	sprintf(text, "-- fs_read: result %d\n", result);
	debug_print(text);

	return result;
}

/*===========================================================================
/	Write the data to the file which is specified through the cookie
/============================================================================*/

int	psionfs_write (void *_ns, void *_node, void *_cookie, off_t pos, const void *buf, size_t *len) {

	dircookie		*dirc = (dircookie*)_cookie;
	nspace*			ns = (nspace*)_ns;	
	vnode*			node = (vnode *)_node;
	char 			text[128];
	int 			result = B_NO_ERROR;
	ssize_t 		ret;
	
	sprintf(text, "## fs_write: vnod_id:%Ld, pos:%Ld, len:%lu %s\n", node->id, pos, *len, node->filename);
	debug_print(text);

	psion_fseek (ns, dirc, SEEK_SET, pos);
	ret = psion_write (ns, dirc, (const char *)buf, *len);
	
	if (ret > 0) {
		*len = ret;
		node->size += ret;		/* we have to remember how much bytes the file is getting bigger	*/
		node->stat_changed = true;
		notify_listener(B_STAT_CHANGED, ns->id, 0, 0, node->id, NULL);	/* Lets tell the system, that there is a new folder!	*/
		sprintf(text, "## fs_write: notify_listener(B_STAT_CHANGED, %ld, 0, 0, %Ld, NULL)\n", ns->id, node->id);
		debug_print(text);

	} else {
		*len = 0;
		switch(ret) {
			case B_NO_ERROR:			result = B_NO_ERROR; break;
			case B_READ_ONLY_DEVICE:	result = B_READ_ONLY_DEVICE; break;
			default:					result = B_FILE_ERROR;	break;
		}
	}
	
	sprintf(text, "-- fs_write: result %d\n", result);
	debug_print(text);

	return result;
}

/*===========================================================================
/	Close the file, specified by the cookie. but don't delete the cookie
/	maybe some buffers have to be cleared
/============================================================================*/

int psionfs_close(void *_ns, void *_node, void *_cookie) {

	nspace*			ns = (nspace*)_ns;	
	vnode*			node = (vnode *)_node;
	dircookie*		cookie = (dircookie*)_cookie;

	char text[128];
	
	sprintf(text, "## fs_close: fd:%d %s\n", cookie->fd, node->filename);
	debug_print(text);
	
	if (node->stat_changed) {
		node->stat_changed = false;
/*		notify_listener(B_STAT_CHANGED, ns->id, node->id, node->id, node->id, NULL);*/	/* Lets tell the system, that there is a new folder!	*/
/*		notify_listener(B_ATTR_CHANGED, ns->id, node->id, node->id, node->id, NULL);*/	/* Lets tell the system, that there is a new folder!	*/
	}
	
	return B_NO_ERROR;
}

/*===========================================================================
/	Delete the cookie and cleanup all the reserved data
/============================================================================*/

int	psionfs_free_cookie(void *_ns, void *node, void *_cookie) {

	char text[128];
	dircookie *cookie = (dircookie*)_cookie;
	nspace* ns = (nspace*)_ns;	

	sprintf(text, "## fs_free_cookie: fd:%d\n", cookie->fd);
	debug_print(text);
	psion_close(ns, cookie);
	
	return B_NO_ERROR;
}

/*===========================================================================
/	we have to unlink the file. first check, if it is a normal file
/============================================================================*/

int	psionfs_unlink(void *_ns, void *_dir, const char *name)
{
	char text[128];
	int			result = B_NO_ERROR;
	vnode* 		dir = (vnode *)_dir;	
	nspace*		ns = (nspace*)_ns;
	vnode*		unlinkNode = NULL;
	vnode_id	vnid;
	
	sprintf(text, "## fs_unlink directory_id:%Ld, filename:%s\n", dir->id, name);
	debug_print(text);
	
	
	unlinkNode = search_vnode(ns, &vnid, dir->id, name, NULL);	/* der node wird gesetzt	*/

	if (unlinkNode == NULL) return B_ENTRY_NOT_FOUND;			/* file not found!			*/
	if (get_vnode(ns->id, vnid, (void**)&unlinkNode) != 0) return B_BAD_VALUE;	/* initialise it on the kernel	*/
 
	if (unlinkNode->mode & S_IFDIR) return B_IS_A_DIRECTORY;	/* want to unlink a directory	*/
	remove_vnode(ns->id, unlinkNode->id);						/* remove the vnod-id			*/
	unlinkNode->deleted = true;									/* set the mark, that the node is deleted	*/
	notify_listener(B_ENTRY_REMOVED, ns->id, dir->id, 0, unlinkNode->id, NULL);	/* Lets tell the system, that there is a file deleted!	*/

	sprintf(text, "## fs_write: notify_listener(B_ENTRY_REMOVED, %ld, %Ld, 0, %Ld, NULL)\n", ns->id, dir->id, unlinkNode->id);
	debug_print(text);

	put_vnode(ns->id, unlinkNode->id);							/* we aren't interested in the data any longer	*/
	return result;
}


/*===========================================================================
/	fill in the stat-structur from a file
/============================================================================*/

int	psionfs_rstat(void *_ns, void *_node, struct stat *st)
{
	vnode*	node = (vnode*)_node;
	nspace* ns = (nspace*)_ns;
	char text[128];
	
	sprintf(text, "## fs_rstat: vnod_id:%Ld, parid:%Ld filename:%s size:%ld\n", node->id, node->parID, node->filename, node->size);
	debug_print(text);

	st->st_dev = ns->id;	/* thats really neccesary that the system is recognising the fs as a device	*/
	st->st_ino = node->id;
	st->st_mode = node->mode;
	st->st_nlink = 1;
	st->st_uid = 0;			/* we have all root rights	*/
	st->st_gid = 0;			/* we have always the master group rights	*/
	st->st_size = node->size;
	st->st_rdev = 0;
	st->st_blksize = MAX_BLOCK;
	st->st_ctime = st->st_mtime = st->st_atime = st->st_crtime = node->modtime;	/* the psion has no different timestamps	*/
	
	return B_NO_ERROR;
}

/*===========================================================================
/	change the stat-structure
/============================================================================*/

/* fs_wstat - fill in stat struct	*/

int	psionfs_wstat(void *_ns, void *_node, struct stat *st, long mask)
{
	vnode*	node = (vnode*)_node;
	nspace* ns = (nspace*)_ns;
	char file[128];
	char text[128];
	int result = B_FILE_ERROR;
	
	find_path(ns, node, file, sizeof(file));	/* find the complete filename	*/

	sprintf(text, "## fs_wstat: vnod_id:%Ld, file:%s mask:%x\n", node->id, file, mask);
	debug_print(text);
	
	switch (mask) {
		case WSTAT_MODE:
			psion_sfstat(ns, file, 0, 0);
			result = B_NO_ERROR;
		case WSTAT_UID:
		case WSTAT_GID:
			result = B_NO_ERROR;	/* we can't change the User or GroupId	but we have to say that we can do this */
			break;
		case WSTAT_SIZE:
			result = B_UNSUPPORTED;	/* not allowed to change the size of the file	*/
			break;
		case WSTAT_ATIME:			/* no change, when there was an accesstime-change	*/
			result = B_UNSUPPORTED;
			break;
		case WSTAT_MTIME:			/* last modificationtime	*/
			psion_sfdate(ns, file, st->st_mtime);
			node->modtime = st->st_mtime;
			result = B_NO_ERROR;
			break;
		case WSTAT_CRTIME:			/* we don't want to change the creationtime!	*/
			result = B_UNSUPPORTED;
			break;
	}
	return result;
}
